import array
from collections import OrderedDict
import asyncio
import datetime as dt
import json
from fastapi import FastAPI
from ib_async import *
import logging
import webbrowser
from loguru import logger
import pytz
from sqlalchemy.orm import Session
from optrabot.broker.brokerconnector import BrokerConnector
from optrabot import schemas
from optrabot.broker.brokerfactory import BrokerFactory
from optrabot.marketdatatype import MarketDataType
from optrabot.optionhelper import OptionHelper
import optrabot.config as optrabotcfg
from optrabot.trademanager import TradeManager
from optrabot.tradetemplate.templatefactory import Template
import optrabot.symbolinfo as symbolInfo
from optrabot.tradetemplate.templatetrigger import TriggerType
from optrabot.tradetemplate.processor.templateprocessor import TemplateProcessor
from optrabot.signaldata import SignalData
from .tradinghubclient import TradinghubClient
from .timesync import TimeSync
import pkg_resources
from .database import *
from . import crud
from apscheduler.schedulers.asyncio import AsyncIOScheduler
from apscheduler.schedulers.base import SchedulerNotRunningError

optrabot_dev_version = '0.18.1-dev'

def get_version() -> str:
	"""
	Returns the version of the package
	"""
	environment = os.environ.get('OPTRABOT_ENV', 'PROD')
	if environment != 'DEV':
		try:
			return pkg_resources.get_distribution('optrabot').version
		except pkg_resources.DistributionNotFound:
			return optrabot_dev_version 
	else:
		return optrabot_dev_version

class OptraBot():
	def __init__(self, app: FastAPI):
		self.app = app
		self._apiKey = None
		self.thc : TradinghubClient = None
		self._marketDataType : MarketDataType = None
		self.Version = get_version()
		self._backgroundScheduler = AsyncIOScheduler()
		self._backgroundScheduler.start()
		self._timeSync : TimeSync = None
			
	def __setitem__(self, key, value):
		setattr(self, key, value)

	def __getitem__(self, key):
		return getattr(self, key)
	
	async def startup(self):
		logger.info('OptraBot {version}', version=self.Version)
		# Read Config
		conf = optrabotcfg.Config("config.yaml")
		optrabotcfg.appConfig = conf
		self['config'] = conf
		conf.logConfigurationData()
		conf.readTemplates()
		conf.readFlows()
		updateDatabase()
		self.thc = TradinghubClient(self)
		if self.thc._apiKey == None:
			return

		try:
			additional_data = {
				'instance_id': conf.getInstanceId(),
				'accounts': self._getConfiguredAccounts()
			}
			await self.thc.connect(additional_data)
		except Exception as excp:
			logger.error('Problem on Startup: {}', excp)
			logger.error('OptraBot halted!')
			return
		
		logger.info('Sucessfully connected to OptraBot Hub')
		
		# OTB-253: createBrokerConnectors() now handles Trade Recovery
		# Trade Recovery runs after first broker connection (see BrokerFactory._onBrokerConnected)
		await BrokerFactory().createBrokerConnectors()
		
		self.thc.start_polling(self._backgroundScheduler)
		TradeManager()

		# Initialize FlowEngine
		from optrabot.flowengine import FlowEngine
		flow_engine = FlowEngine()
		flow_engine.initialize(conf.getFlows(), self._backgroundScheduler)

		self._backgroundScheduler.add_job(self._statusInfo, 'interval', minutes=5, id='statusInfo', misfire_grace_time=None)
		self._backgroundScheduler.add_job(self._new_day_start, 'cron', hour=0, minute=0, second=0, timezone=pytz.timezone('US/Eastern'), id='day_change', misfire_grace_time=None)
		self._backgroundScheduler.add_job(self._check_price_data, 'interval', seconds=30, id='check_price_data', misfire_grace_time=None)
		
		# OTB-258: Initialize time synchronization
		self._timeSync = TimeSync(self.thc)
		# Run immediately at startup
		self._backgroundScheduler.add_job(
			self._timeSync.check_and_sync_time,
			'date',
			run_date=dt.datetime.now() + dt.timedelta(seconds=5),
			id='time_sync_startup',
			misfire_grace_time=None
		)
		# Run daily at 6:00 AM Eastern Time
		self._backgroundScheduler.add_job(
			self._timeSync.check_and_sync_time,
			'cron',
			hour=6,
			minute=0,
			second=0,
			timezone=pytz.timezone('US/Eastern'),
			id='time_sync_daily',
			misfire_grace_time=None
		)
		
		self._scheduleTimeTriggers()

		environment = os.environ.get('OPTRABOT_ENV', 'PROD')
		if environment == 'DEV':
			logger.warning('Run the UI by command: npm start')
		else:
			webPort = conf.get('general.port')
			url = f"http://localhost:{webPort}/ui/optrabot.html"
			self._backgroundScheduler.add_job(self._open_browser, 'date', run_date=dt.datetime.now() + dt.timedelta(seconds=2), args=[url], id='open_browser', misfire_grace_time=None)
	
	async def _open_browser(self, url: str):
		"""
		Opens the default web browser to the specified URL
		"""
		try:
			webbrowser.open(url)
		except Exception as e:
			logger.warning(f'Error opening web browser: {e}')

	async def shutdown(self):
		logger.info('Shutting down OptraBot')
		
		# Shutdown FlowEngine
		try:
			from optrabot.flowengine import FlowEngine
			FlowEngine().shutdown()
		except Exception as e:
			logger.warning(f'Error shutting down FlowEngine: {e}')
		
		# Shutdown TradinghubClient
		try:
			if hasattr(self, 'thc') and self.thc is not None:
				await self.thc.shutdown()
		except Exception as e:
			logger.warning(f'Error shutting down TradinghubClient: {e}')
		
		# Shutdown TradeManager
		try:
			TradeManager().shutdown()
		except Exception as e:
			logger.warning(f'Error shutting down TradeManager: {e}')
		
		# Shutdown BrokerFactory
		try:
			await BrokerFactory().shutdownBrokerConnectors()
		except Exception as e:
			logger.warning(f'Error shutting down BrokerConnectors: {e}')
		
		# Shutdown Background Scheduler
		try:
			if hasattr(self, '_backgroundScheduler') and self._backgroundScheduler is not None:
				try:
					if hasattr(self._backgroundScheduler, 'running') and self._backgroundScheduler.running:
						self._backgroundScheduler.shutdown()
						logger.debug('OptraBot background scheduler shutdown completed')
					else:
						logger.debug('OptraBot background scheduler is not running, skipping shutdown')
				except SchedulerNotRunningError:
					logger.debug('OptraBot background scheduler already stopped')
				except Exception as e:
					logger.warning(f'Error shutting down OptraBot background scheduler: {e}')
		except Exception as e:
			logger.warning(f'Error during OptraBot background scheduler shutdown: {e}')
		
		logger.info('OptraBot shutdown completed')

	async def _check_price_data(self):
		"""
		Checks if the connected brokers are still delivering price data during the trading session
		"""
		await BrokerFactory().check_price_data()

	async def _new_day_start(self):
		"""
		Perform operations on start of a new day
		"""
		logger.debug('Performing Day Change operations')
		await BrokerFactory().new_day_start()
		self._scheduleTimeTriggers()

	def _statusInfo(self):
		siHubConnection = 'OK' if self.thc.isHubConnectionOK() == True else 'Problem!'

		managedTrades = TradeManager().getManagedTrades()
		activeTrades = 0
		for managedTrade in managedTrades:
			if managedTrade.isActive():
				activeTrades += 1

		logger.info(f'Broker Trading enabled: {BrokerFactory().get_trading_satus_info()}')
		logger.info(f'Status Info: Hub Connection: {siHubConnection} - Active Trades: {activeTrades}')

	def _scheduleTimeTriggers(self):
		"""
		Schedules the time triggers for the relevant templates with the time trigger
		"""
		conf: Config = self['config']
		now = dt.datetime.now().astimezone(pytz.UTC)
		for item in conf.getTemplates():
			template : Template = item
			trigger = template.getTrigger()
			if trigger is None:
				continue
			if trigger.type == TriggerType.Time:
				if template.is_enabled() == False:
					logger.debug(f'Template {template.name} is disabled. Not scheduling time trigger.')
					continue
				if trigger._weekdays:
					if now.weekday() not in trigger._weekdays:
						logger.debug(f'Template {template.name} is not scheduled for today')
						continue
				if trigger._blackout_days:
					if now.date() in trigger._blackout_days:
						logger.debug(f'Template {template.name} is in blackout days for today')
						continue
				if not BrokerFactory().is_trading_day():
					logger.debug(f'Market is not open. Not scheduling time trigger for template {template.name}')
					continue
				if trigger._trigger_time.time() < now.time():
					logger.debug(f'Trigger time {trigger._trigger_time} for template {template.name} is in the past. Not scheduling it.')
					continue

				trigger_time_today = dt.datetime.combine(now.date(), trigger._trigger_time.time(), tzinfo=trigger._trigger_time.tzinfo)
				logger.debug(f'Scheduling one-time trigger for template {template.name} at {trigger_time_today}')
				self._backgroundScheduler.add_job(
					self._triggerTemplate,
					'date',
					run_date=trigger_time_today,
					timezone=trigger._trigger_time.tzinfo,
					args=[template],
					id=f'time_trigger_{template.name}',
					misfire_grace_time=None
				)

	async def _triggerTemplate(self, template: Template):
		logger.info(f'Executing Time Trigger for template {template.name}')
		job_id = 'processtemplate' + str(template.name)
		signal_data = SignalData(timestamp=dt.datetime.now().astimezone(pytz.UTC), close=0, strike=0 )
		templateProcessor = TemplateProcessor()
		# Wrap processTemplate call to handle exceptions and send notifications
		self._backgroundScheduler.add_job(
			self._processTemplateWithErrorHandling, 
			args=[templateProcessor, template, signal_data], 
			id=job_id, 
			max_instances=10, 
			misfire_grace_time=None
		)
	
	async def _processTemplateWithErrorHandling(self, templateProcessor, template, signalData):
		"""
		Wrapper for processTemplate that catches exceptions and sends user notifications
		"""
		try:
			await templateProcessor.processTemplate(template, signalData)
		except Exception as e:
			logger.error(f'Template processing failed for {template.name}: {e}')
			from optrabot.tradinghubclient import TradinghubClient, NotificationType
			try:
				await TradinghubClient().send_notification(
					NotificationType.ERROR,
					f'❌ *Template Processing Failed*\n'
					f'*Template:* {template.name}\n'
					f'*Strategy:* {template.strategy}\n'
					f'*Account:* {template.account}\n'
					f'*Error:* {str(e)}'
				)
			except Exception as notify_error:
				logger.error(f'Failed to send error notification: {notify_error}')

	def getMarketDataType(self) -> MarketDataType:
		""" Return the configured Market Data Type
		"""
		if self._marketDataType is None:
			config: Config = self['config']
			try:
				confMarketData = config.get('tws.marketdata')
			except KeyError as keyError:
				confMarketData = 'Delayed'
			self._marketDataType = MarketDataType()
			self._marketDataType.byString(confMarketData)
		return self._marketDataType

	def _getConfiguredAccounts(self) -> list:
		""" 
		Returns a list of configured accounts
		"""
		#conf: Config = self['config']
		conf: Config = optrabotcfg.appConfig
		configuredAccounts = None
		for item in conf.getTemplates():
			template : Template = item
			if configuredAccounts == None:
				configuredAccounts = [template.account]
			else:
				if not template.account in configuredAccounts:
					configuredAccounts.append(template.account)
		return configuredAccounts

	@logger.catch
	def handleTaskDone(self, task: asyncio.Task):
		if not task.cancelled():
			taskException = task.exception()
			if taskException != None:
				logger.error('Task Exception occured!')
				raise taskException


from dataclasses import dataclass
import datetime


@dataclass
class SignalData:
	"""
	Class for holding signal data especially from external signals
	"""
	
	timestamp: datetime
	"""
	Timestamp of the signal
	"""

	close: float
	"""
	Close price of the underlying asset at signal time
	"""

	strike: float
	"""
	Defined strike price of the signal. It may be empty
	"""
	




from datetime import datetime
from typing import List
from loguru import logger
from optrabot.broker.order import Leg, OptionRight, Order, OrderAction, OrderType
from optrabot.signaldata import SignalData
from optrabot.tradetemplate.processor.templateprocessorbase import TemplateProcessorBase
from optrabot.tradetemplate.templatefactory import IronFly, Template
from optrabot.managedtrade import ManagedTrade

class IronFlyProcessor(TemplateProcessorBase):
	
	def __init__(self, template: Template):
		"""
		Initializes the Iron Fly Processor with the given template
		"""
		super().__init__(template)
	
	def composeEntryOrder(self, signalData: SignalData = None):
		"""
		Composes the entry order for the put spread template
		"""
		super().composeEntryOrder(signalData)
		iron_fly_template :IronFly = self._template
		short_strike = None
		# Short Strike Determination
		if signalData and signalData.strike > 0:
			short_strike = signalData.strike

		if short_strike == None:
			raise ValueError('Short Strike could not be determined!')
		
		logger.debug(f'Using Short Strike: {short_strike}')

		if iron_fly_template.wing == None:
			raise ValueError('Wing size has not been defined!')
		logger.debug(f'Using Wing Size: {iron_fly_template.wing}')

		long_strike_upper = short_strike + iron_fly_template.wing
		long_strike_lower = short_strike - iron_fly_template.wing

		# Now create the entry order with its legs as all required strikes are determined
		legs: List[Leg] = []
		legs.append(Leg(action=OrderAction.SELL, strike=short_strike, symbol=self._template.symbol, right=OptionRight.PUT, expiration=self._determined_expiration_date, quantity=1))
		legs.append(Leg(action=OrderAction.SELL, strike=short_strike, symbol=self._template.symbol, right=OptionRight.CALL, expiration=self._determined_expiration_date, quantity=1))
		legs.append(Leg(action=OrderAction.BUY, strike=long_strike_upper, symbol=self._template.symbol, right=OptionRight.CALL, expiration=self._determined_expiration_date, quantity=1))
		legs.append(Leg(action=OrderAction.BUY, strike=long_strike_lower, symbol=self._template.symbol, right=OptionRight.PUT, expiration=self._determined_expiration_date, quantity=1))

		entryOrder = Order(symbol=self._template.symbol, legs=legs, action=OrderAction.BUY_TO_OPEN, quantity=self._template.amount, type=OrderType.LIMIT)
		return entryOrder

	def composeTakeProfitOrder(self, managedTrade: ManagedTrade, fillPrice: float) -> Order:
		"""
		Composes the take profit order based on the template and the given fill price
		"""
		super().composeTakeProfitOrder(managedTrade, fillPrice)
		logger.debug('Creating take profit order for template {}', self._template.name)
		takeProfitPrice = self._template.calculateTakeProfitPrice(fillPrice)
		logger.debug(f'Calculated take profit price: {takeProfitPrice}')

		takeProfitOrder = Order(symbol=self._template.symbol, legs=managedTrade.entryOrder.legs, action=OrderAction.SELL_TO_CLOSE, quantity=self._template.amount, type=OrderType.LIMIT, price=takeProfitPrice)
		return takeProfitOrder
	
	def composeStopLossOrder(self, managedTrade: ManagedTrade, fillPrice: float) -> Order:
		"""
		Composes the stop loss order based on the template and the given fill price
		"""
		super().composeStopLossOrder(managedTrade, fillPrice)
		logger.debug('Creating stop loss order for template {}', self._template.name)
		stopLossPrice = self._template.calculateStopLossPrice(fillPrice)
		logger.debug(f'Calculated stop loss price: {stopLossPrice}')
		
		stopLossOrder = Order(symbol=self._template.symbol, legs=managedTrade.entryOrder.legs, action=OrderAction.SELL_TO_CLOSE, quantity=self._template.amount, type=OrderType.STOP, price=stopLossPrice)
		return stopLossOrder


"""
Flow Event Data Classes

This module defines the event data structures that are passed to flows
when they are triggered.
"""

from dataclasses import dataclass
from datetime import date
from enum import Enum
from typing import Optional


class FlowEventType(str, Enum):
    """Enumeration of available flow event types"""
    EARLY_EXIT = "early_exit"
    TRADE_OPENED = "trade_opened"
    STOP_LOSS_HIT = "stop_loss_hit"
    TAKE_PROFIT_HIT = "take_profit_hit"


@dataclass
class FlowEventData:
    """Base class for flow event data"""
    event_type: FlowEventType
    trade_id: int
    trade_amount: int
    trade_symbol: str
    trade_strategy: str
    template_name: str
    trade_expiration: date
    trade_group_id: str  # Trade group ID for grouping related trades (no default here!)
    
    def get_variables(self) -> dict:
        """
        Returns a dictionary of variables that can be used in flow expressions.
        Variable names use the EVENT-TRADE- prefix.
        """
        return {
            'EVENT_TRADE_ID': self.trade_id,
            'EVENT_TRADE_AMOUNT': self.trade_amount,
            'EVENT_TRADE_SYMBOL': self.trade_symbol,
            'EVENT_TRADE_STRATEGY': self.trade_strategy,
            'EVENT_TRADE_EXPIRATION': self.trade_expiration,
            'EVENT_TRADE_GROUP_ID': self.trade_group_id,
        }


@dataclass
class TradeOpenedEventData(FlowEventData):
    """Event data for trade_opened events"""
    trade_entry_price: float
    
    def __post_init__(self):
        self.event_type = FlowEventType.TRADE_OPENED
    
    def get_variables(self) -> dict:
        variables = super().get_variables()
        variables.update({
            'EVENT_TRADE_ENTRY_PRICE': self.trade_entry_price,
        })
        return variables


@dataclass
class TradeExitEventData(FlowEventData):
    """Base class for trade exit events (early_exit, stop_loss_hit, take_profit_hit)"""
    trade_entry_price: float
    trade_exit_price: float
    trade_net_result: float
    trade_premium: float
    trade_fees: float
    
    def get_variables(self) -> dict:
        variables = super().get_variables()
        variables.update({
            'EVENT_TRADE_ENTRY_PRICE': self.trade_entry_price,
            'EVENT_TRADE_EXIT_PRICE': self.trade_exit_price,
            'EVENT_TRADE_NET_RESULT': self.trade_net_result,
            'EVENT_TRADE_PREMIUM': self.trade_premium,
            'EVENT_TRADE_FEES': self.trade_fees,
        })
        return variables


@dataclass
class EarlyExitEventData(TradeExitEventData):
    """Event data for early_exit events"""
    
    def __post_init__(self):
        self.event_type = FlowEventType.EARLY_EXIT


@dataclass
class StopLossHitEventData(TradeExitEventData):
    """Event data for stop_loss_hit events"""
    
    def __post_init__(self):
        self.event_type = FlowEventType.STOP_LOSS_HIT


@dataclass
class TakeProfitHitEventData(TradeExitEventData):
    """Event data for take_profit_hit events"""
    
    def __post_init__(self):
        self.event_type = FlowEventType.TAKE_PROFIT_HIT



from dataclasses import dataclass
from datetime import date
from loguru import logger
from ib_async import Fill
import re
from optrabot.broker.order import OrderAction
from optrabot import models
from optrabot.tradestatus import TradeStatus

@dataclass
class SecurityStatusData:
	securityId: str
	sec_type: str
	strike: float
	expiration: date
	openContracts: int
	fees: float
	unrealPNL: float
	realPNL: float

class TradeHelper:

	@staticmethod
	def getTradeIdFromOrderRef(orderRef: str) -> int:
		""" Extracts the OptraBot Trade Id from the order reference of a fill
		"""
		tradeId = 0
		pattern = r'^OTB\s\((?P<tradeid>[0-9]+)\):[\s0-9A-Za-z]*'
		compiledPattern = re.compile(pattern)
		match = compiledPattern.match(orderRef)
		if match:
			tradeId = int(match.group('tradeid'))
		return tradeId
	
	@staticmethod
	def isTransactionsComplete(trade: models.Trade, adjustExpired: bool = False) -> bool:
		""" Checks if the transactions of the trade with the given Id are complete,
		which means everything liquidated or expired.
		If 'adjustExpired' is True, then missing EXP transactions are created if the contract is expired already
		"""
		securityStatus = {}
		isComplete = True
		for ta in trade.transactions:
			# Dictionary aus Security-String und dem aktuellen Bestand
			# Am Ende müssen die Bestände 0 sein, damit der Trade Complete ist
			security = ta.sectype + str(ta.expiration) + str(ta.strike)
			change = ta.contracts
			if ta.type == 'SELL':
				change = change * -1
			try:
				currentStatus = securityStatus[security]
				securityStatus[security] = currentStatus+change
			except KeyError:
				securityStatus.update({security:change})
		
		for security, status in securityStatus.items():
			if status != 0:
				isComplete = False
		
		return isComplete
	
	@staticmethod
	def updateTrade(trade: models.Trade, session = None):
		""" Analyzes the transactions of the given trade and 
			- updates the status of the trade
			- calculates the realized PNL
			
		Args:
			trade: The trade to update
			session: Optional SQLAlchemy session to use for database operations.
					 If None, a new session will be created (backward compatible).
		"""
		from optrabot import symbolinfo
		
		# Get symbol multiplier (default to 100 if symbol not found)
		multiplier = 100
		if trade.symbol in symbolinfo.symbol_infos:
			multiplier = symbolinfo.symbol_infos[trade.symbol].multiplier
		else:
			logger.warning(f'Symbol {trade.symbol} not found in symbol_infos, using default multiplier 100')
		
		securityStatus = {}
		isComplete = True
		for ta in trade.transactions:
			security = ta.sectype + str(ta.expiration) + str(ta.strike)
			change = ta.contracts
			taFee = ta.fee + ta.commission
			# OTB-262: Case-insensitive transaction type comparison
			# Database has 'Buy', 'Sell' (from OrderAction enum values)
			# But need to handle legacy or broker-specific variations
			tx_type = ta.type
			is_sell = (tx_type == OrderAction.SELL or 
					   tx_type == 'Sell' or 
					   (isinstance(tx_type, str) and tx_type.upper() == 'SELL'))
			if is_sell:
				change = change * -1
			try:
				statusData = securityStatus[security]
				statusData.openContracts += change
			except KeyError:
				statusData = SecurityStatusData(securityId=security, sec_type=ta.sectype, strike=ta.strike, expiration=ta.expiration, openContracts=change, fees=ta.fee, unrealPNL=0, realPNL = 0)
				securityStatus.update({security:statusData})
			statusData.fees += taFee
			# Track unrealized P&L (price impact + fees)
			statusData.unrealPNL -= ((ta.price * multiplier * change) + taFee)
		
		# For EXPIRED trades: Create closing transactions for open positions
		if trade.status == TradeStatus.EXPIRED:
			logger.debug(f'Trade {trade.id} is EXPIRED - checking for open positions requiring closing transactions')
			from optrabot import crud
			from optrabot.database import SessionLocal
			import datetime
			import pytz
			
			# Use provided session or create a new one
			use_existing_session = session is not None
			db_session = session if use_existing_session else SessionLocal()
			
			try:
				max_tx_id = crud.getMaxTransactionId(db_session, trade.id)
				exp_transactions_created = 0
				
				for security, statusData in securityStatus.items():
					if statusData.openContracts != 0:
						# Position is open - create closing transaction at $0 (expired worthless)
						# If openContracts > 0: Long position → SELL to close
						# If openContracts < 0: Short position → BUY to close
						closing_type = OrderAction.SELL if statusData.openContracts > 0 else OrderAction.BUY
						
						max_tx_id += 1
						exp_transaction = models.Transaction(
							tradeid=trade.id,
							id=max_tx_id,
							type=closing_type,
							sectype=statusData.sec_type,
							timestamp=datetime.datetime.now(pytz.UTC),
							expiration=statusData.expiration,
							strike=statusData.strike,
							contracts=abs(statusData.openContracts),
							price=0.0,  # Expired worthless
							fee=0.0,
							commission=0.0,
							notes=f'Auto-generated closing transaction for expired {statusData.sec_type} position'
						)
						db_session.add(exp_transaction)
						exp_transactions_created += 1
						logger.debug(f'Created {closing_type} transaction for Trade {trade.id}: {statusData.sec_type}@{statusData.strike} ({abs(statusData.openContracts)} contracts) - expired worthless')
						
						# OTB-262: Update statusData to reflect the closed position
						# The closing transaction at $0 doesn't change the P&L (already in unrealPNL)
						statusData.openContracts = 0
						statusData.realPNL = statusData.unrealPNL
				
				if exp_transactions_created > 0:
					# Only commit if we created our own session
					if not use_existing_session:
						db_session.commit()
					logger.debug(f'Created {exp_transactions_created} closing transactions for Trade {trade.id}')
					
					# OTB-262: After creating closing transactions, all positions are now closed
					# Force isComplete to True so realizedPNL calculation happens below
					# Note: We already updated statusData.openContracts = 0 for all positions above
					isComplete = True
					logger.debug(f'Trade {trade.id}: All positions closed via expiration, isComplete=True')
				else:
					# No closing transactions were created - positions were already closed
					logger.debug(f'Trade {trade.id}: No closing transactions needed - all positions already closed')
			finally:
				# Only close session if we created it
				if not use_existing_session:
					db_session.close()
					# Check if trade is actually complete by examining securityStatus
					logger.debug(f'Trade {trade.id}: No closing transactions created (positions already closed or no open positions found)')
		
		# Calculate realized P&L based on final position status
		trade.realizedPNL = 0
		for security, statusData in securityStatus.items():
			if statusData.openContracts == 0:
				# Leg is closed - realized P&L is the full unrealized P&L
				statusData.realPNL = statusData.unrealPNL
				trade.realizedPNL += statusData.realPNL
			else:
				# Leg still open - only fees are realized (as costs), no P&L yet
				statusData.realPNL = -statusData.fees  # Negative because fees are costs
				# Don't add to trade.realizedPNL - open positions don't contribute to realized P&L
				# Trade is not complete yet
				# OTB-262: Only set isComplete to False if not already set by EXPIRED logic
				if trade.status != TradeStatus.EXPIRED:
					isComplete = False
		
		if isComplete == True:
			trade.realizedPNL = round(trade.realizedPNL, 2)
			logger.debug(f'Trade {trade.id}: Calculated realizedPNL = ${trade.realizedPNL}')
			# Only set status to CLOSED if it's not already EXPIRED
			# This preserves the distinction between regularly closed trades and expired trades
			if trade.status != TradeStatus.EXPIRED:
				trade.status = TradeStatus.CLOSED




			
				


import subprocess
import sys

def build_with_frontend():
	"""
	Build the OptraBot but including the build of the frontend.
	"""
	try:
		# Schritt 1: Frontend-Build ausführen
		print("Running frontend build...")
		subprocess.run(["npm", "run", "build"], check=True, cwd="./frontend")
		
		# Schritt 2: Poetry-Build ausführen
		print("Running Poetry build...")
		subprocess.run(["poetry", "build"], check=True)

		print("Build completed successfully!")
	except subprocess.CalledProcessError as e:
		print(f"Error during build: {e}")
		return


from datetime import datetime
from enum import Enum
from optrabot.optionhelper import OptionHelper
from typing import Optional

class OrderAction(str, Enum):
    """
    Represents actions of an order or order le
    """

    BUY_TO_OPEN = "Buy to Open"
    BUY_TO_CLOSE = "Buy to Close"
    SELL_TO_OPEN = "Sell to Open"
    SELL_TO_CLOSE = "Sell to Close"
    BUY = "Buy"
    SELL = "Sell"

class OrderType(str, Enum):
	"""
	Represents order types
	"""
	LIMIT = "Limit"
	MARKET = "Market"
	STOP = "Stop"
	STOP_LIMIT = "Stop Limit"

class OrderStatus(str, Enum):
	"""
	Represents the status of an order
	"""
	OPEN = "Open"
	FILLED = "Filled"
	CANCELLED = "Cancelled"

class OptionRight(str, Enum):
	"""
	Represents the right of an option
	"""
	CALL = "Call"
	PUT = "Put"

class Leg():
	"""
	Represents a leg of an order.
	"""
	action: OrderAction
	quantity: Optional[int]
	symbol: str
	strike: float
	right: OptionRight
	expiration: datetime
	askPrice: float = None
	bidPrice: float = None
	midPrice: float = None
	delta: float = None
	brokerSpecific = {}
     
	def __init__(self, action: OrderAction, symbol: str, quantity: Optional[int] = 0, strike: float = None, right: OptionRight = None, expiration: datetime = None):
		self.action = action
		self.quantity = quantity
		self.symbol = symbol
		self.right = right
		self.strike = strike
		self.expiration = expiration
		self.delta = 0

class PriceEffect(str, Enum):
    """
    This is an :class:`~enum.Enum` that shows the sign of a price effect.
    """
    CREDIT = "Credit"
    DEBIT = "Debit"
    NONE = "None"

class Order():
	"""
	Represents an order. A order can have multiple legs
	"""
	
	def __init__(self, symbol: str = '', action: OrderAction = None, type: OrderType = None, legs: list[Leg] = None, quantity: int = 1, price: float = None):
		self.symbol = symbol
		self.action = action
		self.legs: list[Leg] = legs
		self.type = type
		self.quantity = quantity
		self.price: float = price
		self.averageFillPrice: float = None
		self.ocaGroup = None
		self.status: OrderStatus = None
		self.orderReference: str = ''
		self.brokerSpecific = {}
		self.broker_order_id = ''
		self.price_effect = PriceEffect.NONE

	def determine_price_effect(self):
		""" 
		Determines the price effect of the order based on its legs and the price data
		"""
		total_price = self.calculate_price()
		self.price_effect = PriceEffect.CREDIT if total_price < 0 else PriceEffect.DEBIT if total_price > 0 else PriceEffect.NONE

	def calculate_price(self) -> float:
		"""
		Calculates the total price of the order based on its legs and their prices
		"""
		total_price = 0.0
		for leg in self.legs:
			if leg.strike is None or leg.right is None:
				continue  # Skip legs without strike or right
			if leg.midPrice is None:
				leg.midPrice = self.calculate_mid_price(leg.askPrice, leg.bidPrice)
			
			if leg.action in [OrderAction.BUY_TO_OPEN, OrderAction.BUY_TO_CLOSE, OrderAction.BUY]:
				total_price += leg.midPrice * (leg.quantity if leg.quantity else 1)
			elif leg.action in [OrderAction.SELL_TO_OPEN, OrderAction.SELL_TO_CLOSE, OrderAction.SELL]:
				total_price -= leg.midPrice * (leg.quantity if leg.quantity else 1)
		
		return total_price

	def calculate_mid_price(self, ask_price: float, bid_price: float) -> float:
		"""
		Calculates the mid price based on the given ask and bid prices
		"""
		int_ask_price = ask_price if not OptionHelper.isNan(ask_price) and ask_price >= 0 else 0
		int_bid_price = bid_price if not OptionHelper.isNan(bid_price) and bid_price >= 0 else 0
		return (int_ask_price + int_bid_price) / 2

class Execution():
	"""
	Represents a fill information of an order execution
	"""
	def __init__(self, id: str, action: OrderAction, sec_type: str, strike: float, amount: int, price: float, timestamp: datetime, expiration: datetime):
		self.id: str = id
		self.action: OrderAction =action
		self.sec_type: str = sec_type
		self.strike: float = strike
		self.timestamp: datetime = timestamp
		self.expiration: datetime = expiration
		self.amount: int = amount
		self.price: float = price